from functools import reduce
from operator import add
from .consts import OPS
from klefki.bitcoin.sign import sign_bytes


def pay_to_pubhash(privkey, pubkey, address):
    scriptPubKey = reduce(add, [
        OPS.OP_DUP,
        OPS.OP_HASH160,
        address,
        OPS.OP_EQUALVERIFY,
        OPS.OP_CHECKSIG
    ])
    scriptSig = sign_bytes(privkey, scriptPubKey) + pubkey
    return scriptPubKey, scriptSig


def pay_to_scripthash(scripthash):
    return reduce(add, [
    ])



from typing import Iterable, Callable
from functools import reduce


__all__ = ['Isomorphism', 'bijection', 'do']


class Isomorphism():
    '''
    A morphism f : X → Y in a category is an isomorphism if it admits a two-sided inverse.
    '''

    class Trunks(list):

        def __call__(self, *args, **kwargs):
            def _eval(args, fn):
                if type(args) in [list, tuple]:
                    return fn(*args)
                if type(args) is dict:
                    return fn(**args)
                else:
                    return fn(args)

            return reduce(_eval, self, kwargs or args)

        def __invert__(self):
            return self.inverse()

        def __rshift__(self, next):
            self.append(next)
            return self

        def __lshift__(self, prev):
            self.insert(0, prev.inverse)
            return self

        def inverse(self):
            return self.__class__([fn.inverse for fn in self[::-1]])

    def __init__(self, fn):
        assert hasattr(fn, '__call__')
        assert hasattr(fn, 'inverse')
        self.fn = fn
        self.inverse = fn.inverse

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __rshift__(self, next):
        return self.Trunks([self, next])

    def __lshift__(self, prev):
        return self.Trunks([self.inverse, prev.inverse])

    @staticmethod
    def bijection(inverse):
        def _(fn):
            inverse.inverse = fn
            fn.inverse = inverse
            return fn
        return _

    @classmethod
    def do(cls, *fns: Iterable[Callable]) -> 'Trunks':
        return cls.Trunks([cls(fn) for fn in fns])

    def __repr__(self):
        return self.fn.__repr__()


bijection = Isomorphism.bijection
do = Isomorphism.do



from typing import Tuple
import random
from klefki.utils import to_sha256int
from klefki.types.algebra.concrete import (
    JacobianGroupSecp256r1 as JG,
    FiniteFieldSecp256r1 as F,
    EllipticCurveCyclicSubgroupSecp256r1 as CG,
    EllipticCurveGroupSecp256r1 as ECG,
    FiniteFieldCyclicSecp256r1 as CF
)

__all__ = [
    'random_privkey',
    'pubkey',
    'sign',
    'verify',
    'verify_msghash'

]

N = CG.N
G = CG.G
A = ECG.A
B = ECG.B
P = F.P

SigType = Tuple[CF, CF, CF]


def random_privkey() -> CF:
    return CF(random.randint(1, N))


def pubkey(priv: CF) -> ECG:
    return ECG(JG(G @ priv))


def sign(priv: CF, m: str) -> SigType:
    '''
    https://bitcoin.stackexchange.com/questions/38351/ecdsa-v-r-s-what-is-v
    '''
    k = CF(random_privkey())
    z = CF(to_sha256int(m))

    p = G @ k
    r, y = CF(p.value[0]), CF(p.value[1])
    s = (z + priv * r) / k
    v = CF(27 + y.value % 2)
    return v, r, s


def verify(pub: ECG, sig: tuple, msg: str):
    mhash = to_sha256int(msg)
    return verify_msghash(pub, sig, mhash)


def verify_msghash(pub: ECG, sig: tuple, mhash: int):
    if len(sig) == 2:
        r, s = sig
    else:
        v, r, s = sig
    z = CF(mhash)
    u1 = z / s
    u2 = r / s
    rp = G @ u1 + pub @ u2
    return r == rp.value[0]


def recover(sig: tuple, mhash: int):
    v, r, s = sig
    x = F(r)
    xcubedaxb = x ** 3 + F(A) * x + F(B)
    beta = pow(xcubedaxb.value, (P + 1) // 4, P)
    if v.value % 2 ^ beta % 2:
        y = beta
        print('case 1')
    else:
        print('case 2')
        y = P - beta
    Gz = G @ (F(N) - F(mhash))
    Xy = ECG((F(x), F(y))) @ s
    Qr = Gz + Xy
    Q = Qr @ ~CF(r)
    return Q


def recover_via_msg(sig: tuple, msg: str):
    return recover(sig, to_sha256int(msg))


def proof():
    priv = random_privkey()
    m = 'test'
    k = CF(random_privkey())
    z = CF(to_sha256int(m))
    r = CF((G @ k).value[0])
    s = z / k + priv * r / k

    assert k == z / s + priv * r / s
    assert G @ k == G @ (z / s + priv * r / s)
    assert G @ k == G @ (z / s) + G @ priv @ (r / s)

    pub = G @ priv
    assert pub == pubkey(priv)
    assert G @ k == G @ (z / s) + pub @ (r / s)
    u1 = z / s
    u2 = r / s
    assert G @ k == G @ u1 + pub @ u2



SECP256K1_P = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
SECP256K1_N = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
SECP256K1_Gx = 0X79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798
SECP256K1_Gy = 0X483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8
SECP256K1_A = 0
SECP256K1_B = 7

SECP256R1_P = 0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff
SECP256R1_A = 0xffffffff00000001000000000000000000000000fffffffffffffffffffffffc
SECP256R1_B = 0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b
SECP256R1_N = 0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551
SECP256R1_Gx = 0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296
SECP256R1_Gy = 0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5



from .groups import EllipticCurveGroup, CyclicGroup, JacobianGroup
from .fields import FiniteField
import klefki.const as const

__all__ = [
    'FiniteFieldSecp256k1',
    'FiniteFieldCyclicSecp256k1',
    'EllipticCurveGroupSecp256k1',
    'JacobianGroupSecp256k1',
    'EllipticCurveCyclicSubgroupSecp256k1',
    'FiniteFieldSecp256r1',
    'FiniteFieldCyclicSecp256r1',
    'EllipticCurveGroupSecp256r1',
    'JacobianGroupSecp256r1',
    'EllipticCurveCyclicSubgroupSecp256r1'

]


class FiniteFieldSecp256k1(FiniteField):
    __slots__ = ()
    P = const.SECP256K1_P


class FiniteFieldCyclicSecp256k1(FiniteField):
    __slots__ = ()
    P = const.SECP256K1_N


class EllipticCurveGroupSecp256k1(EllipticCurveGroup):
    __slots__ = ()
    A = const.SECP256K1_A
    B = const.SECP256K1_B


class JacobianGroupSecp256k1(JacobianGroup):
    __slots__ = ()
    A = const.SECP256K1_A
    B = const.SECP256K1_B


class EllipticCurveCyclicSubgroupSecp256k1(CyclicGroup):
    __slots__ = ()
    G = EllipticCurveGroupSecp256k1(
        (
            FiniteFieldSecp256k1(const.SECP256K1_Gx),
            FiniteFieldSecp256k1(const.SECP256K1_Gy)
        )
    )
    N = const.SECP256K1_N


class FiniteFieldSecp256r1(FiniteField):
    __slots__ = ()
    P = const.SECP256R1_P


class FiniteFieldCyclicSecp256r1(FiniteField):
    __slots__ = ()
    P = const.SECP256R1_N


class EllipticCurveGroupSecp256r1(EllipticCurveGroup):
    __slots__ = ()
    A = const.SECP256R1_A
    B = const.SECP256R1_B


class JacobianGroupSecp256r1(JacobianGroup):
    __slots__ = ()
    A = const.SECP256R1_A
    B = const.SECP256R1_B


class EllipticCurveCyclicSubgroupSecp256r1(CyclicGroup):
    __slots__ = ()
    G = EllipticCurveGroupSecp256r1(
        (
            FiniteFieldSecp256r1(const.SECP256R1_Gx),
            FiniteFieldSecp256r1(const.SECP256R1_Gy)
        )
    )
    N = const.SECP256R1_N



import hashlib
from klefki.crypto.ecdsa.secp256k1 import pubkey
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    FiniteFieldCyclicSecp256k1 as CF,
)
from klefki.utils import ripemd160, b58encode


from .public import to_bytes


def gen_address(pub: ECG):
    pub = to_bytes(pub)
    pub_sha = hashlib.sha512(pub).digest()
    addy = ripemd160(pub_sha)
    return 'EOS' + b58encode(addy + ripemd160(addy)[:4])


def gen_address_from_priv(priv: CF) -> str:
    return gen_address(pubkey(priv))



from functools import wraps
from .consts import CONSTS


class Stack(object):

    ops = {}

    def __init__(self, script):
        self.storage = []
        self.script = script
        self._output = []

    @classmethod
    def register(self, op):
        def _(fn):
            @wraps(fn)
            def wrapper(*args, **kwargs):
                self.ops[op] = fn
                return fn(*args, **kwargs)
            return wrapper
        return _

    def eval(self):
        return self.ops.get(self.script[0])(self, self.script)

    @property
    def output(self):
        return self._output

    @output.setter
    def set_output(self, v):
        self._output.append(v)

    def isEmpty(self):
        return len(self.storage) == 0

    def push(self, p):
        self.storage[:0] = p
        return self

    def pop(self, p):
        return self.pop(p)

    def __repr__(self):
        return self.storage.__repr__()


@Stack.register(CONSTS.OP_0)
def op_0(stack, script):
    stack.push([])
    stack.ouput = ''
    return stack.eval(script[1], stack, script[1:])


def op_pushdata(stack, script, n):
    next_n_bytes = script[0:n]
    count = int(next_n_bytes.hex(), 16)
    stack.push(script[n: count + n])
    stack.output = next_n_bytes
    return stack.eval(script[count + n], stack, script[count + n:])


@Stack.register(CONSTS.OP_PUSHDATA1)
def op_pushdata1(stack, script):
    return op_pushdata(stack, script, 1)


@Stack.register(CONSTS.OP_PUSHDATA2)
def op_pushdata2(stack, script):
    return op_pushdata(stack, script, 2)


@Stack.register(CONSTS.OP_PUSHDATA3)
def op_pushdata3(stack, script):
    return op_pushdata(stack, script, 3)


def op_pushdata4(stack, script):
    return op_pushdata(stack, script, 4)


def op_1negate(stack, script):
    stack.push(-1)
    stack.output = -1
    return stack.eval(script[1], stack, script[1:])


def op_1(stack, script):
    stack.push(1)
    stack.output = 1
    return stack.eval(script[1], stack, script[1:])



from abc import abstractproperty
from .abstract import Group
from .fields import FiniteField


class EllipticCurveGroup(Group):
    # for y^2 = x^3 + A * x + B
    A = abstractproperty()
    B = abstractproperty()

    def fmap(self, o):
        if isinstance(o, JacobianGroup):
            z = ~(o.value[2])
            return (o.value[0] * z ** 2, o.value[1] * z ** 3)
        return o

    def op(self, g):
        # https://www.desmos.com/calculator/ialhd71we3
        if g.value == 0:
            return self
        if self.value == 0:
            return g

        field = self.value[0].__class__

        if self.value[0] != g.value[0]:
            #  m = \frac{y_1 - y_2}{x_1 - x_2}
            m = (self.value[1] - g.value[1]) / (self.value[0] - g.value[0])

        if self.value[0] == g.value[0]:
            if self.value[1] == -(g.value[1]):
                return self.identity

            m = (field(3) * self.value[0] * self.value[0] +
                 field(self.A)) / (field(2) * self.value[1])

        r_x = (m * m - self.value[0] - g.value[0])
        r_y = (self.value[1] + m * (r_x - self.value[0]))
        return self.__class__((r_x, -r_y))

    def inverse(self):
        if self.value == 0:
            return self
        return self.__class__((self.value[0], -self.value[1]))

    @property
    def identity(self):
        # The abstract zero of EC Group
        return self.__class__(0)


class CyclicGroup(Group):
    '''
    With Lagrange's therem
    the order of a subgroup is a divisor of the order of the parent group
    '''
    # The Base Point
    G = abstractproperty()
    # Order of subgroup
    N = abstractproperty()

    def fmap(self, o):
        if isinstance(o, FiniteField):
            return o.__class__(o.value % self.N)
        return getattr(o, 'value', o) % self.N

    def op(self, g):
        '''
        2 + 3 -> 2G + 3G -> 5G
        '''
        if g.value == 0:
            return self
        res = self.value + g.value
        return self.__class__(res % self.N)

    def inverse(self):
        if self.value == 0:
            return self
        return self.__class__(self.N - 1 - self.value)

    @property
    def identity(self):
        return self.__class__(0)


class JacobianGroup(Group):
    A = abstractproperty()
    B = abstractproperty()

    def fmap(self, o):
        if isinstance(o, EllipticCurveGroup):
            field = o.value[0].__class__
            return [o.value[0], o.value[1], field(1)]
        return o

    def double(self, n=None):
        if not n:
            n = self
        field = n.value[0].__class__
        if not n.value[1].value:
            return n.__class__(
                (
                    field(0),
                    field(0),
                    field(0)
                )
            )

        ysq = self.value[1] ** 2
        S = self.value[0] @ 4 * ysq
        M = (self.value[0] ** 2) @ 3 + (self.value[2] ** 4) @ self.A
        nx = M ** 2 - S @ 2
        ny = M * (S - nx) - (ysq ** 2) @ 8
        nz = field(2) * self.value[1] * self.value[2]
        return self.__class__((nx, ny, nz))

    @property
    def identity(self):
        return self.__class__(0)

    def inverse(self):
        pass

    def op(self, g):
        field = self.value[0].__class__
        if not g.value:
            return self
        if not self.value:
            return g
        U1 = self.value[0] * g.value[2] ** 2
        U2 = g.value[0] * self.value[2] ** 2
        S1 = self.value[1] * g.value[2] ** 3
        S2 = g.value[1] * self.value[2] ** 3
        if U1 == U2:
            if S1 != S2:
                return self.__class__(
                    (
                        field(0),
                        field(0),
                        field(1)
                    )
                )
            return self.double()
        H = field(U2.value - U1.value)
        R = field(S2.value - S1.value)
        H2 = H * H
        H3 = H * H2
        U1H2 = U1 * H2
        nx = (R ** 2) - H3 - (U1H2 @ 2)

        ny = R * (U1H2 - nx) - S1 * H3
        nz = H * self.value[2] * g.value[2]
        return self.__class__((nx, ny, nz))



from klefki.crypto.ecdsa.secp256k1 import random_privkey
from .public import gen_pub_key
from .private import encode_privkey
from .address import gen_address_from_priv

__all__ = ['gen_address_from_priv']


def gen_key_pair():
    key = random_privkey()
    return encode_privkey(key), gen_pub_key(key)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


install_requires = [
    line
    for line in open(
        os.path.join(here, "requirements.txt"),
        "r"
    )
]
author = 'Ryan Kung'
email = 'ryankung@ieee.org'


setup(
    name='klefki',
    description='Klefki is a playground for researching elliptic curve group based cryptocoins, such as Bitcoin and Ethereum. All data types & structures are based on mathematical defination of abstract algebra.',  # noqa
    url='https://github.com/RyanKung/klefki',
    version='0.0.4.0',
    packages=find_packages(here, exclude=['tests']),
    license='GPL',
    author=author,
    author_email=email,
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        "Programming Language :: Python :: 3.5",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    entry_points={
        'console_scripts': [
            'klefki=klefki.client.shell:main'
        ]
    }
)



from typing import Callable
from hashlib import sha256
from klefki.utils import b58encode, int_to_byte

from klefki.crypto.ecdsa.secp256k1 import (
    random_privkey,
)
from klefki.types.algebra.concrete import (
    FiniteFieldCyclicSecp256k1 as CF,
)

from klefki.types.algebra.isomorphism import bijection, do


def wrap_key(key: bytes, version=128, compress=1):
    priv = bytes([version]) + key + bytes([compress])
    auth = sha256(sha256(priv).digest()).digest()[:4]
    res = priv + auth
    assert len(res) == 1 + 32 + 1 + 4
    return res


@bijection(wrap_key)
def unwrap_key(key: bytes):
    return key[1: -5]


@bijection(CF)
def from_cf(a: CF) -> int:
    return a.value


encode_privkey: Callable[[CF], str] = do(
    from_cf,
    int_to_byte,
    wrap_key,
    b58encode
)

decode_privkey: Callable[[str], CF] = ~encode_privkey


def gen_random_privkey():
    return encode_privkey(random_privkey())



from abc import abstractproperty
from .abstract import Field
from klefki.algorithms import extended_euclidean_algorithm


class FiniteField(Field):

    P = abstractproperty()

    def fmap(self, o):
        return getattr(o, 'value', o) % self.P

    @property
    def identity(self):
        return self.__class__(0 % self.P)

    @property
    def sec_identity(self):
        return self.__class__(1 % self.P)

    def inverse(self):
        return self.__class__(self.P - self.value)

    def mod(self, a, b):
        if isinstance(a, complex):
            return complex((a.real % b), (a.imag % b))
        return a % b

    def sec_inverse(self):
        gcd, x, y = extended_euclidean_algorithm(self.value, self.P)
        assert (self.value * x + self.P * y) == gcd
        if gcd != 1:
            # Either n is 0, or p is not prime number
            raise ValueError(
                '{} has no multiplicate inverse '
                'modulo {}'.format(self.value, self.P)
            )
        return self.__class__(x % self.P)

    def op(self, g):
        return self.__class__(
            self.mod(
                (self.value + g.value), self.P
            )
        )

    def sec_op(self, g):
        return self.__class__(
            self.mod(
                (self.value * g.value), self.P
            )
        )



from hashlib import sha256
from klefki.utils import dhash256, ripemd160, b58encode
from klefki.crypto.ecdsa.secp256k1 import (
    pubkey
)
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    EllipticCurveCyclicSubgroupSecp256k1 as CG,
    FiniteFieldSecp256k1 as F
)

__all__ = [
    'gen_address'
]


G = CG.G
A = ECG.A
B = ECG.B
P = F.P


def gen_address(pub: ECG) -> str:
    x, y = pub.value[0].value, pub.value[1].value
    if y % 2 == 0:
        prefix = bytes([0x02])
    else:
        prefix = bytes([0x03])
    networkid = bytes([0x00])
    hashed = ripemd160(
        sha256(
            prefix + x.to_bytes(32, byteorder='big')
        ).digest()
    )
    assert len(hashed) == 20
    with_network = networkid + hashed
    auth = dhash256(with_network)[:4]
    res = with_network + auth
    assert len(res) == 25
    return b58encode(res)


def gen_address_from_priv(key) -> str:
    return gen_address(pubkey(key))



import hashlib
from typing import Callable
from klefki.utils import b58encode, int_to_byte
from klefki.crypto.ecdsa.secp256k1 import (
    random_privkey,
)
from klefki.types.algebra.concrete import (
    FiniteFieldCyclicSecp256k1 as CF,
)

from klefki.types.algebra.isomorphism import bijection, do


@bijection(CF)
def from_cf(a: CF) -> int:
    return a.value


def add_version(a: bytes) -> bytes:
    return bytes([0x80]) + a


@bijection(add_version)
def remove_version(a: bytes) -> bytes:
    _a = bytearray(a)
    assert hex(_a[0]) == '0x80'
    return _a[1:]


def add_checksum(a: bytes) -> bytes:
    assert len(a) == 33
    return a + hashlib.sha256(hashlib.sha256(a).digest()).digest()[:4]


@bijection(add_checksum)
def remove_checksum(a: bytes) -> bytes:
    res = bytes(bytearray(a)[:-4])
    checksum = bytes(bytearray(a)[-4:])
    new_checksum = hashlib.sha256(hashlib.sha256(res).digest()).digest()[:4]
    assert checksum == new_checksum
    return res


encode_privkey: Callable[[CF], str] = do(
    from_cf,
    int_to_byte,
    add_version,
    add_checksum,
    b58encode
)

decode_privkey: Callable[[str], CF] = ~encode_privkey


def gen_random_privkey():
    return encode_privkey(random_privkey())



from klefki.utils import int_to_byte
from klefki.crypto.ecdsa.secp256k1 import pubkey
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    FiniteFieldCyclicSecp256k1 as CF

)
from klefki.types.algebra.isomorphism import bijection, do
from klefki.utils import ripemd160, b58encode, byte_to_int


def to_bytes(pub: ECG) -> str:
    return bytes([2 + (pub.value[1].value % 2)]) + int_to_byte(pub.value[0].value)

@bijection(to_bytes)
def from_bytes(a):
    return ECG([
        CF(byte_to_int(a[:1]) - 2),
        CF(byte_to_int(a[1:]))
    ])


def checksum(a) -> int:
    assert len(a) == 33
    return 'EOS' + b58encode(a + ripemd160(a)[:4])

@bijection(checksum)
def unchecksum(a):
    return b58encode.inverse(a[3:])


def gen_pub_key(key: CF):
    return checksum(to_bytes(pubkey(key)))
